PROTOCOL_VERSION = 10

# PROTOCOL_VERSION 10 (6/8/2020)
# change maximum transaction count in BDN performance stats message
#
# PROTOCOL_VERSION 9 (4/17/2020)
# add broadcast_type to broadcast messages
#
# PROTOCOL_VERSION 8 (2/5/2020)
# add timestamp to tx message types
#
# PROTOCOL_VERSION 7
# added TxQuotaFlag to transaction messages
